"""PLUGIN_NAME: PLUGIN_DESC

Describe your plugin here and do whatever you want. It should at least have a
__version__ attribute to specify the version of the package
"""

__version__ = "PLUGIN_VERSION"
